import argparse
import inspect
import logging
import os
import sys

from .__main__ import setup_logging

log = logging.getLogger()

os.environ["NUMEXPR_MAX_THREADS"] = "8"

DESCRIPTION = """
hic-internal is used for automation, not intend to be used by end user. 
Use hicluster instead. 
"""

EPILOG = ''


def impute_chromosome_internal_subparser(subparser):
    parser = subparser.add_parser('impute-chromosome',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="RWR imputation for one chromosome in one cell")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--scool_url",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--chrom",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--output_path",
        type=str,
        required=True
    )

    parser.add_argument(
        '--logscale',
        dest='logscale',
        action='store_true')
    parser.set_defaults(logscale=False)

    parser.add_argument(
        "--pad",
        type=int,
        default=1
    )

    parser.add_argument(
        "--std",
        type=int,
        default=1
    )

    parser.add_argument(
        "--rp",
        type=float,
        default=0.5
    )

    parser.add_argument(
        "--tol",
        type=float,
        default=0.01
    )

    parser.add_argument(
        "--window_size",
        type=int,
        default=500000000
    )

    parser.add_argument(
        "--step_size",
        type=int,
        default=10000000
    )

    parser.add_argument(
        "--output_dist",
        type=int,
        default=500000000
    )

    parser.add_argument(
        "--min_cutoff",
        type=float,
        default=0
    )
    return


def aggregate_chromosomes_internal_subparser(subparser):
    parser = subparser.add_parser('aggregate-chromosomes',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Aggregate chromosome HDFs for one cell")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--chrom_size_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--input_dir",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--chrom_wildcard",
        type=str,
        default='{chrom}.hdf'
    )

    parser.add_argument(
        '--csr',
        dest='csr',
        action='store_true')
    parser.set_defaults(csr=False)


def calculate_loop_matrix_internal_subparser(subparser):
    parser = subparser.add_parser('calculate-loop-matrix',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Calculate Loop Matrix E and T for single chromosome of one cell")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--cell_url",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--chrom",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--output_prefix",
        type=str,
        required=True
    )

    parser.add_argument(
        "--dist",
        type=int,
        default=5050000
    )

    parser.add_argument(
        "--cap",
        type=int,
        default=5
    )

    parser.add_argument(
        "--pad",
        type=int,
        default=5
    )

    parser.add_argument(
        "--gap",
        type=int,
        default=2
    )

    parser.add_argument(
        "--min_cutoff",
        type=float,
        default=1e-6
    )

    parser.add_argument(
        '--shuffle',
        dest='shuffle',
        action='store_true',
        help='Shuffle E for background calculation'
    )
    parser.set_defaults(shuffle=False)

    parser.add_argument(
        '--log_e',
        dest='log_e',
        action='store_true',
        help='Normalize E at log scale'
    )
    parser.set_defaults(log_e=False)


def merge_cell_impute_matrix_internal_subparser(subparser):
    parser = subparser.add_parser('merge-cell-impute-matrix',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Merge Loop Matrix E and T from cells to group")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--cell_urls_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--chrom",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_prefix",
        type=str,
        required=True
    )
    return


def merge_cell_loop_bkg_internal_subparser(subparser):
    parser = subparser.add_parser('merge-loop-matrix',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Merge Loop Matrix E and T from cells to group")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--output_dir",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_prefix",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--merge_type",
        type=str,
        required=True
    )


def merge_group_chunks_internal_subparser(subparser):
    parser = subparser.add_parser('merge-group-chunks',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Aggregate group chunks cool files to one scool file per group.")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--chrom_size_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--group",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_dir",
        type=str,
        required=True
    )

    parser.add_argument(
        "--matrix_types",
        type=str,
        nargs='+',
        default=['E', 'E2', 'T', 'T2', 'Q'],
        required=False
    )


def merge_raw_scool_internal_subparser(subparser):
    parser = subparser.add_parser('merge-raw-scool',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Merge single cell raw matrix by cluster stored in scool files.")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--chrom_size_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--cell_table_path",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--output_dir",
        type=str,
        required=True
    )

    parser.add_argument(
        "--cpu",
        type=int,
        default=1
    )

    parser.add_argument(
        '--add_trans',
        dest='add_trans',
        action='store_true',
        help='Whether add trans matrix also.'
    )
    parser.set_defaults(add_trans=False)


def call_loop_internal_subparser(subparser):
    parser = subparser.add_parser('call-loop',
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                  help="Call loop using group matrix stored in scool file.")
    parser_req = parser.add_argument_group("Required inputs")

    parser_req.add_argument(
        "--group_prefix",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--resolution",
        type=int,
        required=True
    )

    parser_req.add_argument(
        "--output_prefix",
        type=str,
        required=True
    )

    parser_req.add_argument(
        "--thres_bl",
        type=float,
        default=1.33
    )

    parser_req.add_argument(
        "--thres_donut",
        type=float,
        default=1.33
    )

    parser_req.add_argument(
        "--thres_h",
        type=float,
        default=1.2
    )

    parser_req.add_argument(
        "--thres_v",
        type=float,
        default=1.2
    )

    parser_req.add_argument(
        "--fdr_thres",
        type=float,
        default=0.1
    )

    parser_req.add_argument(
        "--dist_thres",
        type=float,
        default=20000
    )

    parser_req.add_argument(
        "--size_thres",
        type=float,
        default=1
    )


def internal_main():
    parser = argparse.ArgumentParser(description=DESCRIPTION,
                                     epilog=EPILOG,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)
    subparsers = parser.add_subparsers(
        title="functions",
        dest="command",
        metavar=""
    )

    # add subparsers
    current_module = sys.modules[__name__]
    # get all functions in parser
    for name, register_subparser_func in inspect.getmembers(current_module, inspect.isfunction):
        if 'internal_subparser' in name:
            register_subparser_func(subparsers)

    # initiate
    args = None
    if len(sys.argv) > 1:
        args = parser.parse_args()
    else:
        # print out help
        parser.parse_args(["-h"])
        exit()

    # set up logging
    if not logging.root.handlers:
        setup_logging(stdout=True,
                      quiet=False)

    # execute command
    args_vars = vars(args)
    for k, v in args_vars.items():
        logging.debug(k, v, type(v), sep='\t')

    cur_command = args_vars.pop('command')
    # Do real import here:
    if cur_command == 'impute-chromosome':
        from .impute.impute_chromosome import impute_chromosome as func
    elif cur_command == 'aggregate-chromosomes':
        from .cool.utilities import aggregate_chromosomes as func
    elif cur_command == 'calculate-loop-matrix':
        from .loop.loop_bkg import calculate_chrom_background_normalization as func
    elif cur_command == 'merge-loop-matrix':
        from .loop.merge_cell_to_group import merge_cells_for_single_chromosome as func
    elif cur_command == 'merge-group-chunks':
        from .loop.merge_cell_to_group import merge_group_chunks_to_group_cools as func
    elif cur_command == 'merge-cell-impute-matrix':
        from .impute.merge_cell_to_group import merge_cells_for_single_chromosome as func
    elif cur_command == 'call-loop':
        from .loop.loop_calling import call_loops as func
    elif cur_command == 'merge-raw-scool':
        from .loop.merge_raw_matrix import merge_raw_scool_by_cluster as func
    else:
        log.debug(f'{cur_command} not Known, check the main function if else part')
        parser.parse_args(["-h"])
        return

    # run the command
    func(**args_vars)
    return



import pathlib
import time
import numpy as np
from scipy.sparse import csr_matrix, load_npz, triu
from ..cool import write_coo, get_chrom_offsets
from .merge_cell_to_group import read_single_cool_chrom
import cooler
import pandas as pd
import h5py
from concurrent.futures import ProcessPoolExecutor, as_completed

def chrom_ave_iterator(chunk_dirs,
                       chrom_sizes,
                       chrom_offset,
                       matrix_type,
                       total_cells):
    print(f'Reading matrix {matrix_type}')
    for chrom in chrom_sizes.keys():
        # sum together multiple chunks
        # first
        cool_path = list(chunk_dirs[0].glob(f'*/*.{matrix_type}.cool'))[0]
        with h5py.File(cool_path, 'a') as f:
            n_cells = f.attrs['group_n_cells']
        matrix = read_single_cool_chrom(cool_path, chrom) * n_cells
        # others
        for chunk_dir in chunk_dirs[1:]:
            cool_path = list(chunk_dir.glob(f'*/*.{matrix_type}.cool'))[0]
            with h5py.File(cool_path, 'a') as f:
                n_cells = f.attrs['group_n_cells']
            matrix += read_single_cool_chrom(cool_path, chrom) * n_cells

        matrix = matrix.tocoo()
        pixel_df = pd.DataFrame({
            'bin1_id': matrix.row,
            'bin2_id': matrix.col,
            'count': matrix.data
        })
        pixel_df.iloc[:, :2] += chrom_offset[chrom]
        # All the chunk
        pixel_df.iloc[:, -1] /= total_cells
        yield pixel_df


def save_single_matrix_type(cooler_path,
                            bins_df,
                            chunk_dirs,
                            chrom_sizes,
                            chrom_offset, 
                            matrix_type,
                            total_cells):
    chrom_iter = chrom_ave_iterator(chunk_dirs,
                                    chrom_sizes,
                                    chrom_offset,
                                    matrix_type,
                                    total_cells)
    cooler.create_cooler(cool_uri=cooler_path,
                         bins=bins_df,
                         pixels=chrom_iter,
                         ordered=True,
                         dtypes={'count': np.float32})
    with h5py.File(cooler_path, 'a') as f:
        f.attrs['group_n_cells'] = total_cells
    return


def merge_group_to_bigger_group_cools(chrom_size_path,
                                      resolution,
                                      group,
                                      output_dir,
                                      group_list,
                                      shuffle,
                                      matrix_types=('E', 'E2', 'T', 'T2', 'Q')):
    """
    Sum all the chunk sum cool files,
    and finally divide the total number of cells to
    get a group cell number normalized cool file in the end.
    """
    # determine chunk dirs for the group:
    output_dir = pathlib.Path(output_dir).absolute()
    group_dir = output_dir / group
    group_dir.mkdir(exist_ok=True, parents=True)

    group_list = [pathlib.Path(xx) for xx in group_list]
    
    # count total cells
    total_cells = 0
    for chunk_dir in group_list:
        total_cells += pd.read_csv(chunk_dir / 'cell_table.tsv', index_col=0).shape[0]
    
    if shuffle:
        group_list = [xx / 'shuffle/' for xx in group_list]

    chrom_sizes = cooler.read_chromsizes(chrom_size_path, all_names=True)
    bins_df = cooler.binnify(chrom_sizes, resolution)
    chrom_offset = get_chrom_offsets(bins_df)

    with ProcessPoolExecutor(5) as exe:
        futures = {}
        for matrix_type in matrix_types:
            cooler_path = str(group_dir / f'{group}.{matrix_type}.cool')
            future = exe.submit(save_single_matrix_type,
                                cooler_path=cooler_path,
                                bins_df=bins_df,
                                chunk_dirs=group_list,
                                chrom_sizes=chrom_sizes,
                                chrom_offset=chrom_offset,
                                matrix_type=matrix_type,
                                total_cells=total_cells)
            futures[future] = matrix_type
        for future in as_completed(futures):
            matrix_type = futures[future]
            print(f'Matrix {matrix_type} generated')
            future.result()
    return





import pandas as pd
import numpy as np
import pathlib
import cooler
import h5py
import shutil
from concurrent.futures import ProcessPoolExecutor, as_completed

from ..cool import get_chrom_offsets
from .merge_cell_to_group import read_single_cool_chrom


def _chrom_sum_iterator(cell_urls,
                        chrom_sizes,
                        chrom_offset,
                        add_trans=False):
    """
    Iterate through the raw matrices and chromosomes of cells.

    Parameters
    ----------
    cell_urls :
        List of cell urls.
    chrom_sizes :
        Dictionary of chromosome sizes.
    chrom_offset :
        Dictionary of chromosome offsets.
    add_trans :
        If true, will also iterate all the trans combinations (different chromosomes).

    Yields
    -------
    pixel_df :
        Dataframe of pixels. Used by cooler.create_cooler to save to h5 file.
    """

    def _iter_1d(_chrom1, _chrom2):
        # sum together multiple chunks
        # first
        cell_url = cell_urls[0]
        matrix = read_single_cool_chrom(cell_url, chrom=_chrom1, chrom2=_chrom2)

        # others
        if len(cell_urls) > 1:
            for cell_url in cell_urls[1:]:
                matrix += read_single_cool_chrom(cell_url, chrom=_chrom1, chrom2=_chrom2)

        matrix = matrix.tocoo()
        _pixel_df = pd.DataFrame({
            'bin1_id': matrix.row,
            'bin2_id': matrix.col,
            'count': matrix.data
        })
        if _chrom2 is None:
            # both row and col are chrom1
            _pixel_df.iloc[:, :2] += chrom_offset[_chrom1]
        else:
            # row is chrom1, add chrom1 offset
            _pixel_df.iloc[:, 0] += chrom_offset[_chrom1]
            # col is chrom2, add chrom2 offset
            _pixel_df.iloc[:, 1] += chrom_offset[_chrom2]
        return _pixel_df

    if add_trans:
        # only iter upper triangle
        # chrom order by offset, small to large
        chroms = [k for k, v in sorted(chrom_offset.items(), key=lambda i: i[1])]
        n_chroms = len(chroms)
        for a in range(n_chroms):
            chrom1 = chroms[a]
            chrom1_dfs = []
            for b in range(a, n_chroms):
                chrom2 = chroms[b]
                pixel_df = _iter_1d(chrom1, chrom2)
                chrom1_dfs.append(pixel_df)
            chrom1_df = pd.concat(chrom1_dfs).sort_values(by=['bin1_id', 'bin2_id'])
            yield chrom1_df
    else:
        for chrom in chrom_sizes.keys():
            pixel_df = _iter_1d(chrom, None)
            yield pixel_df


def _save_single_matrix_type(cooler_path,
                             bins_df,
                             cell_urls,
                             chrom_sizes,
                             chrom_offset,
                             add_trans=False):
    """
    Save a single matrix type Cool file from merging multiple cell urls.

    Parameters
    ----------
    cooler_path :
        Path to the output cool file.
    bins_df :
        Dataframe of bins. Created from chromosome sizes and resolution.
    cell_urls :
        List of cell urls to merge.
    chrom_sizes :
        Dictionary of chromosome sizes.
    chrom_offset :
        Dictionary of chromosome offsets.
    add_trans :
        Whether add trans matrix also.
    """
    chrom_iter = _chrom_sum_iterator(cell_urls,
                                     chrom_sizes,
                                     chrom_offset,
                                     add_trans=add_trans)
    cooler.create_cooler(cool_uri=cooler_path,
                         bins=bins_df,
                         pixels=chrom_iter,
                         ordered=True,
                         dtypes={'count': np.float32})
    with h5py.File(cooler_path, 'a') as f:
        f.attrs['group_n_cells'] = len(cell_urls)
    return


def make_raw_matrix_cell_table(cell_table, resolution_str='10K'):
    # get all the raw matrix cell url automatically if the dir path is the default
    scool_dirs = cell_table['cell_url'].apply(lambda i: '/'.join(i.split('/')[:-4])).unique()
    cell_urls = {}
    scool_file_pattern = f'raw/*.{resolution_str}.scool'
    for scool_dir in scool_dirs:
        for scool_path in pathlib.Path(scool_dir).glob(scool_file_pattern):
            with h5py.File(scool_path, 'r') as _cool:
                cell_ids = list(_cool['cells'].keys())
                for cell_id in cell_ids:
                    cell_urls[cell_id] = f'{scool_path}::/cells/{cell_id}'
    raw_url_series = pd.Series(cell_urls)

    # delete old url
    del cell_table['cell_url']
    # add new url
    cell_table['cell_url'] = raw_url_series

    na_cells = cell_table['cell_url'].isna().sum()
    if na_cells > 0:
        raise ValueError(f'{na_cells} cells do not have raw matrix.')

    cell_table = cell_table[['cell_url', 'cell_group']]
    return cell_table


def merge_raw_scool_by_cluster(chrom_size_path, resolution, cell_table_path,
                               output_dir, add_trans=False, cpu=1):
    """
    Sum the raw matrix of cells, no normalization.

    Parameters
    ----------
    chrom_size_path :
        Path to the chrom size file. This file is used to determine chromosome names and bins.
    resolution :
        Resolution of the raw matrix.
    cell_table_path :
        Path to the cell table.
        This table should contain three columns: cell_id, cell_url, cell_group; no Header.
        The cell_id is the id of the cell in the raw matrix.
        The cell_url is the path to the raw matrix.
        The cell_group is the group of the cells.
    output_dir :
        Path to the output directory. Group cool files will be named as "<output_dir>/<cell_group>.cool".
    add_trans :
        Whether add trans matrix also.
    cpu :
        Number of CPUs to use.

    """
    # determine chunk dirs for the group:
    output_dir = pathlib.Path(output_dir).absolute()
    output_dir.mkdir(exist_ok=True)
    cell_table = pd.read_csv(cell_table_path,
                             sep='\t',
                             index_col=0,
                             header=None,
                             names=['cell_id', 'cell_url', 'cell_group'])

    chrom_sizes = cooler.read_chromsizes(chrom_size_path, all_names=True)
    bins_df = cooler.binnify(chrom_sizes, resolution)
    chrom_offset = get_chrom_offsets(bins_df)

    with ProcessPoolExecutor(cpu) as exe:
        futures = {}
        for cell_group, sub_df in cell_table.groupby('cell_group'):
            cell_urls = sub_df['cell_url'].tolist()
            cooler_path = output_dir / f'{cell_group}.cool'
            if cooler_path.exists():
                print(f'{cooler_path} already exists, skip.')
                continue

            cooler_temp_path = str(output_dir / f'{cell_group}.temp.cool')
            future = exe.submit(_save_single_matrix_type,
                                cooler_path=cooler_temp_path,
                                bins_df=bins_df,
                                cell_urls=cell_urls,
                                chrom_sizes=chrom_sizes,
                                chrom_offset=chrom_offset,
                                add_trans=add_trans)
            futures[future] = cell_group
        for future in as_completed(futures):
            cell_group = futures[future]
            print(f'Matrix {cell_group} generated')
            future.result()

            # move the temp cool file to the final location
            cooler_path = str(output_dir / f'{cell_group}.cool')
            cooler_temp_path = str(output_dir / f'{cell_group}.temp.cool')
            shutil.move(cooler_temp_path, cooler_path)
    return



import pathlib
import shutil
from concurrent.futures import ProcessPoolExecutor, as_completed

import cooler
import numpy as np
import pandas as pd
import xarray as xr
import zarr
from cooler.util import read_chromsizes, binnify
from numcodecs import Blosc

SMALL_SAMPLE_CHUNK = 1
COMPRESSOR_C_LEVEL = 3


class CoolDSSingleMatrixWriter:
    def __init__(self,
                 path,
                 cool_table_path,
                 value_types,
                 chrom_sizes_path,
                 chrom1,
                 chrom2=None,
                 mode='w',
                 cooler_bin_size=10000,
                 bin_chunk_size=510,
                 sample_chunk_size=50,
                 data_dtype='float32',
                 cpu=1):
        """
        Write a single chrom1-by-chrom2 matrix to CoolDS zarr.

        Parameters
        ----------
        path :
            Path to the zarr dir.
        cool_table_path :
            Path to the cool table.
        value_types :
            Dict of cool types and their value types.
        chrom_sizes_path :
            Path to the chrom sizes file.
        chrom1 :
            Chrom1 name.
        chrom2 :
            Chrom2 name. If None, chrom1 will be used.
        mode :
            Mode to open the zarr.
        cooler_bin_size :
            Cooler bin size.
        bin_chunk_size :
            Chunk size of the bin1 and bin2 dimensions.
        sample_chunk_size :
            Chunk size of the sample dimension.
        data_dtype :
            Data type of the matrix.
        cpu :
            Number of CPUs to use.
        """
        self.path = path
        self.root = zarr.open(path, mode=mode)
        self.log_dir_path = None
        self.mode = mode
        self.value_types = value_types
        self.cool_tables, self.sample_ids = self._read_cool_table(cool_table_path)
        self.n_sample = self.sample_ids.size
        self.n_cpu = cpu

        self.bin_chunk_size = bin_chunk_size
        self.sample_chunk_size = sample_chunk_size
        self.data_dtype = data_dtype

        self.cooler_bin_size = cooler_bin_size
        self.chrom_sizes, self.chrom_bins, self.n_bins = self._read_chrom_info(chrom_sizes_path)

        self.chrom1 = chrom1
        if chrom2 is None:
            self.chrom2 = chrom1
        else:
            self.chrom2 = chrom2

        self.chrom1_size = self.chrom_sizes[self.chrom1]
        self.chrom2_size = self.chrom_sizes[self.chrom2]
        self.chrom1_bins = self.chrom_bins[self.chrom_bins['chrom'] == self.chrom1]
        self.chrom2_bins = self.chrom_bins[self.chrom_bins['chrom'] == self.chrom2]
        self.chrom1_n_bins = self.chrom1_bins.shape[0]
        self.chrom2_n_bins = self.chrom2_bins.shape[0]

        self.execute()

    def _read_cool_table(self, cool_table_path):
        value_types = self.value_types
        # get cool tables for each cool type
        cool_paths = pd.read_csv(cool_table_path,
                                 header=None,
                                 names=['sample', 'value_type', 'path', 'cool_type'])

        _cool_tables = {}
        sample_ids = []
        for cool_type, sub_paths in cool_paths.groupby('cool_type'):
            cool_table = sub_paths.pivot(index='sample',
                                         columns='value_type',
                                         values='path')

            assert cool_type in value_types, f'cool_type "{cool_type}" not in value_types provided by user.'
            value_type_list = value_types[cool_type]
            assert all([value in cool_table.columns for value in value_type_list
                        ]), f'Some {cool_type} values in {value_type_list} missing in cool_table'
            assert cool_table.isna().values.sum(
            ) == 0, f'{cool_type} cool path table has some missing paths.'

            _cool_tables[cool_type] = cool_table
            sample_ids += cool_table.index.tolist()

        sample_ids = pd.Index(set(sample_ids))
        cool_tables = {}
        for cool_type, table in _cool_tables.items():
            assert table.index.size == sample_ids.size, 'sample id not consistent between cool types'
            cool_tables[cool_type] = table.loc[sample_ids].copy()

        return cool_tables, sample_ids

    def _read_chrom_info(self, chrom_sizes_path):
        # chrom length and bins info
        chrom_sizes = read_chromsizes(chrom_sizes_path)
        chrom_bins = binnify(chrom_sizes, self.cooler_bin_size)
        n_bins = chrom_bins.shape[0]
        return chrom_sizes, chrom_bins, n_bins

    def _add_root_attrs(self):
        self.root.attrs['chrom1'] = self.chrom1
        self.root.attrs['chrom2'] = self.chrom2
        self.root.attrs['chrom1_size'] = self.chrom1_size
        self.root.attrs['chrom2_size'] = self.chrom2_size
        self.root.attrs['chrom1_n_bins'] = self.chrom1_n_bins
        self.root.attrs['chrom2_n_bins'] = self.chrom2_n_bins
        self.root.attrs['cooler_bin_size'] = self.cooler_bin_size
        self.root.attrs['bin_chunk_size'] = self.bin_chunk_size
        self.root.attrs['sample_chunk_size'] = self.sample_chunk_size

    def _init_zarr(self):
        root = zarr.open(self.path, mode=self.mode)

        # add log dir
        self.log_dir_path = pathlib.Path(self.path) / '.log'
        self.log_dir_path.mkdir(exist_ok=True, parents=True)

        # prepare flag
        success_flag = self.log_dir_path / "_PREPARE_SUCCESS"
        if success_flag.exists():
            print("Zarr structure already initiated. Skipping...")

        # create sample_id
        sample_id = root.require_dataset("sample_id",
                                         shape=(self.n_sample,),
                                         chunks=(self.n_sample,),
                                         dtype="<U256")
        sample_id[:] = list(self.sample_ids)
        sample_id.attrs["_ARRAY_DIMENSIONS"] = "sample_id"

        # change compressor
        # according to https://zarr.readthedocs.io/en/stable/tutorial.html#compressors
        # and my test, zstd has 1.6x compression ratio compared to default blosc lz4
        compressor = Blosc(cname='zstd', clevel=COMPRESSOR_C_LEVEL, shuffle=Blosc.SHUFFLE)

        # create empty da
        for cool_type, value_type_list in self.value_types.items():
            # value type index
            n_value_type = len(value_type_list)
            value_dim = f"{cool_type}_value_type"
            value_type_da = root.require_dataset(value_dim,
                                                 shape=(n_value_type,),
                                                 chunks=(n_value_type,),
                                                 dtype="<U10")
            value_type_da[:] = list(value_type_list)
            value_type_da.attrs["_ARRAY_DIMENSIONS"] = [value_dim]

            # data
            z = root.require_dataset(
                cool_type,
                shape=(self.chrom1_n_bins, self.chrom2_n_bins, self.n_sample, n_value_type),
                chunks=(self.bin_chunk_size, self.bin_chunk_size, self.sample_chunk_size, 1),
                dtype=self.data_dtype,
                compressor=compressor
            )
            z.attrs["_ARRAY_DIMENSIONS"] = ["bin1", "bin2", "sample_id", value_dim]

        # add root attrs
        self._add_root_attrs()

        # consolidate metadata
        zarr.consolidate_metadata(self.path)

        # touch success flag
        success_flag.touch()
        return

    def _save_cool_to_temp_zarr(self, cool_paths, output_path, cool_type, value_type):
        da_list = []
        for sample, cool_path in cool_paths.items():
            cool = cooler.Cooler(cool_path)
            matrix = cool.matrix(balance=False, sparse=False).fetch(self.chrom1, self.chrom2).astype(self.data_dtype)
            assert matrix.shape == (self.chrom1_n_bins, self.chrom2_n_bins)

            # keep only the upper triangle of the matrix
            matrix = np.triu(matrix)

            # da.dims {'bin1': n_bins, 'bin2': n_bins, 'sample_id': 1, 'value_type': 1}
            da = xr.DataArray(matrix, dims=['bin1', 'bin2']).chunk({
                'bin1': self.bin_chunk_size,
                'bin2': self.bin_chunk_size
            }).expand_dims(dim={
                'sample_id': [sample],
                f'{cool_type}_value_type': [value_type]
            }).transpose('bin1', 'bin2', 'sample_id', f'{cool_type}_value_type')
            da_list.append(da)
        da = xr.concat(da_list, dim='sample_id')
        ds = xr.Dataset({cool_type: da})
        ds.to_zarr(output_path, mode='w')
        return

    def _save_small_sample_chunks(self):
        with ProcessPoolExecutor(self.n_cpu) as executor:
            futures = {}
            for cool_type, cool_table in self.cool_tables.items():
                for value_type, sample_cool_path in cool_table.items():
                    for chunk_start in range(0, self.n_sample, SMALL_SAMPLE_CHUNK):
                        sample_chunk = self.sample_ids[chunk_start:chunk_start + SMALL_SAMPLE_CHUNK]
                        cool_paths = sample_cool_path.loc[sample_chunk]
                        output_path = f'{self.chrom1}_{self.chrom2}_{cool_type}_{value_type}_{chunk_start}_temp.zarr'
                        future = executor.submit(self._save_cool_to_temp_zarr,
                                                 cool_paths=cool_paths,
                                                 output_path=output_path,
                                                 cool_type=cool_type,
                                                 value_type=value_type
                                                 )
                        futures[future] = [cool_type, value_type, chunk_start, output_path]

            temp_zarr_records = []
            for future in as_completed(futures):
                cool_type, value_type, chunk_start, output_path = futures[future]
                try:
                    future.result()
                except Exception as exc:
                    print(f"Cool type {cool_type} value_type {value_type} "
                          f"chunk_start {chunk_start} generated an exception: {exc}")
                    raise exc
                else:
                    temp_zarr_records.append([cool_type, value_type, chunk_start, output_path])

        temp_zarr_records = pd.DataFrame(temp_zarr_records, columns=['cool_type',
                                                                     'value_type',
                                                                     'chunk_start',
                                                                     'output_path'])
        return temp_zarr_records

    @staticmethod
    def _save_single_bin_chunk_worker(ds_paths,
                                      zarr_path,
                                      cool_type,
                                      bin1_slice,
                                      bin2_slice,
                                      sample_id_slice,
                                      value_idx):
        ds = xr.open_mfdataset(ds_paths, concat_dim='sample_id', combine='nested', engine='zarr')
        zarr_da = zarr.open(zarr_path, mode="r+")
        # ds only contains one value type
        _data = ds[cool_type][bin1_slice, bin2_slice, sample_id_slice, 0].load()
        if (_data != 0).sum() == 0:
            return

        zarr_da[bin1_slice, bin2_slice, sample_id_slice, value_idx] = _data
        return

    def _save_single_bin_chunk(self, zarr_path, path_array, cool_type):
        with ProcessPoolExecutor(self.n_cpu) as executor:
            futures = {}
            for value_idx, (_, paths) in enumerate(path_array.items()):
                ds = xr.open_mfdataset(paths, concat_dim='sample_id', combine='nested', engine='zarr')
                bin1_idx = ds.get_index('bin1')
                bin2_idx = ds.get_index('bin2')
                sample_id_idx = ds.get_index('sample_id')
                for bin1_start in range(0, bin1_idx.size, self.bin_chunk_size):
                    for bin2_start in range(0, bin2_idx.size, self.bin_chunk_size):
                        for sample_id_start in range(0, sample_id_idx.size, self.sample_chunk_size):
                            bin1_slice = slice(bin1_start, bin1_start + self.bin_chunk_size)
                            bin2_slice = slice(bin2_start, bin2_start + self.bin_chunk_size)
                            sample_id_slice = slice(sample_id_start, sample_id_start + self.sample_chunk_size)
                            future = executor.submit(self._save_single_bin_chunk_worker,
                                                     ds_paths=paths.tolist(),
                                                     zarr_path=zarr_path,
                                                     cool_type=cool_type,
                                                     bin1_slice=bin1_slice,
                                                     bin2_slice=bin2_slice,
                                                     sample_id_slice=sample_id_slice,
                                                     value_idx=value_idx)
                            futures[future] = (bin1_start, bin2_start, sample_id_start)

            for future in as_completed(futures):
                bin1_start, bin2_start, sample_id_start = futures[future]
                try:
                    future.result()
                except Exception as e:
                    print(f'Got error when saving bin1 {bin1_start} to {bin1_start + self.bin_chunk_size} '
                          f'bin2 {bin2_start} to {bin2_start + self.bin_chunk_size} '
                          f'sample_id {sample_id_start} to {sample_id_start + self.sample_chunk_size}')
                    raise e
        return

    def _save_small_sample_chunks_ds_to_final_zarr(self, temp_zarr_records):
        for cool_type, path_df in temp_zarr_records.groupby('cool_type'):
            # make sure the temp zarr paths are in the correct order
            path_array = path_df.pivot(
                index='chunk_start', columns='value_type', values='output_path'
            ).sort_index()[self.value_types[cool_type]]

            zarr_path = f'{self.path}/{cool_type}/'
            self._save_single_bin_chunk(zarr_path=zarr_path,
                                        path_array=path_array,
                                        cool_type=cool_type)
        return

    def _write_data(self):
        # config blosc compressor when run multi-processing
        # see zarr doc here: https://zarr.readthedocs.io/en/stable/tutorial.html#configuring-blosc
        if self.n_cpu > 1:
            from numcodecs import blosc
            blosc.use_threads = False

        final_success_flag = self.log_dir_path / "_WRITE_SUCCESS"
        if final_success_flag.exists():
            print("Data already written. Skipping...")

        # first, convert cool to temp zarr,
        # where the whole matrix is loaded into memory,
        # and sample_id chunk size is small (to limit memory usage)
        temp_zarr_records = self._save_small_sample_chunks()

        # second, write temp zarr to final zarr,
        # where the matrix is chunked into small chunks (self.bin_chunk_size, to limit memory usage),
        # and sample_id chunk size is final (self.sample_chunk_size)
        self._save_small_sample_chunks_ds_to_final_zarr(temp_zarr_records=temp_zarr_records)

        # remove temp zarr files
        for _, row in temp_zarr_records.iterrows():
            shutil.rmtree(row['output_path'])

        # touch success file
        final_success_flag.touch()
        return

    def execute(self):
        """Execute the pipeline."""
        self._init_zarr()
        self._write_data()
        return


def generate_cool_ds(output_dir,
                     cool_table_path,
                     value_types,
                     chrom_sizes_path,
                     trans_matrix=False,
                     mode='w',
                     cooler_bin_size=10000,
                     bin_chunk_size=510,
                     sample_chunk_size=50,
                     data_dtype='float32',
                     cpu=1):
    """
    Generate a CoolDS zarr dataset from cool files.

    Parameters
    ----------
    output_dir :
        The output directory.
    cool_table_path :
        The path to the cool table.
    value_types :
        Dict of cool types and their value types.
    chrom_sizes_path :
        Path to the chrom sizes file.
    trans_matrix :
        Whether generate trans-contacts (chrom1 != chrom2) matrix
    mode :
        Mode to open the zarr.
    cooler_bin_size :
        Cooler bin size.
    bin_chunk_size :
        Chunk size of the bin1 and bin2 dimensions.
    sample_chunk_size :
        Chunk size of the sample dimension.
    data_dtype :
        Data type of the matrix.
    cpu :
        Number of CPUs to use.
    """
    output_dir = pathlib.Path(output_dir).absolute().resolve()
    output_dir.mkdir(parents=True, exist_ok=True)

    chrom_sizes = read_chromsizes(chrom_sizes_path)

    for chrom1 in chrom_sizes.index:
        for chrom2 in chrom_sizes.index:
            if not trans_matrix:
                if chrom1 != chrom2:
                    continue
            if chrom1 == chrom2:
                path = output_dir / chrom1
            else:
                path = output_dir / f'{chrom1}-{chrom2}'

            CoolDSSingleMatrixWriter(path,
                                     cool_table_path=cool_table_path,
                                     value_types=value_types,
                                     chrom_sizes_path=chrom_sizes_path,
                                     chrom1=chrom1,
                                     chrom2=chrom2,
                                     mode=mode,
                                     cooler_bin_size=cooler_bin_size,
                                     bin_chunk_size=bin_chunk_size,
                                     sample_chunk_size=sample_chunk_size,
                                     data_dtype=data_dtype,
                                     cpu=cpu)
    return
